@@ -10,6 +10,7 @@
 #include <setjmp.h>				/* setjmp, longjmp */
 #include <stdio.h>				/* snprintf */
 #include <stdarg.h>				/* va_* */
+#include <stdbool.h>				/* bool */
 #include <stdlib.h>				/* free */
 #include <string.h>				/* str* */
 #include <unistd.h>				/* getpagesize */
@@ -25,7 +26,7 @@
 #include <libipset/transport.h>			/* transport */
 #include <libipset/mnl.h>			/* default backend */
 #include <libipset/utils.h>			/* STREQ */
-#include <libipset/ui.h>			/* IPSET_ENV_* */
+#include <libipset/ipset.h>			/* IPSET_ENV_* */
 #include <libipset/session.h>			/* prototypes */
 
 #define IPSET_NEST_MAX	4
@@ -43,15 +44,19 @@ struct ipset_session {
 	const struct ipset_type *saved_type;	/* Saved type */
 	struct nlattr *nested[IPSET_NEST_MAX];	/* Pointer to nest levels */
 	uint8_t nestid;				/* Current nest level */
+	uint8_t protocol;			/* The protocol used */
 	bool version_checked;			/* Version checked */
 	/* Output buffer */
 	char outbuf[IPSET_OUTBUFLEN];		/* Output buffer */
 	enum ipset_output_mode mode;		/* Output mode */
-	ipset_outfn outfn;			/* Output function */
+	ipset_print_outfn print_outfn;		/* Output function to file */
+	void *p;				/* Private data for print_outfn */
+	/* Session IO */
+	bool normal_io, full_io;		/* Default/normal/full IO */
+	FILE *istream, *ostream;		/* Session input/output stream */
 	/* Error/warning reporting */
 	char report[IPSET_ERRORBUFLEN];		/* Error/report buffer */
-	char *errmsg;
-	char *warnmsg;
+	enum ipset_err_type err_type;		/* ERROR/WARNING/NOTICE */
 	uint8_t envopts;			/* Session env opts */
 	/* Kernel message buffer */
 	size_t bufsize;
@@ -115,41 +120,24 @@ ipset_session_lineno(struct ipset_session *session, uint32_t lineno)
 	session->lineno = lineno;
 }
 
+/**
+ * ipset_session_printf_private - returns the session private pointer
+ * @session: session structure
+ *
+ * Returns the private pointer in the session structure,
+ * for private/custom print fuctions.
+ */
+void *
+ipset_session_printf_private(struct ipset_session *session)
+{
+	assert(session);
+	return session->p;
+}
+
 /*
  * Environment options
  */
 
-/**
- * ipset_envopt_parse - parse/set environment option
- * @session: session structure
- * @opt: environment option
- * @arg: option argument (unused)
- *
- * Parse and set an environment option.
- *
- * Returns 0 on success or a negative error code.
- */
-int
-ipset_envopt_parse(struct ipset_session *session, int opt,
-		   const char *arg UNUSED)
-{
-	assert(session);
-
-	switch (opt) {
-	case IPSET_ENV_SORTED:
-	case IPSET_ENV_QUIET:
-	case IPSET_ENV_RESOLVE:
-	case IPSET_ENV_EXIST:
-	case IPSET_ENV_LIST_SETNAME:
-	case IPSET_ENV_LIST_HEADER:
-		session->envopts |= opt;
-		return 0;
-	default:
-		break;
-	}
-	return -1;
-}
-
 /**
  * ipset_envopt_test - test environment option
  * @session: session structure
@@ -166,6 +154,34 @@ ipset_envopt_test(struct ipset_session *session, enum ipset_envopt opt)
 	return session->envopts & opt;
 }
 
+/**
+ * ipset_envopt_set - set environment option
+ * @session: session structure
+ * @opt: environment option
+ *
+ * Set an environment option of the session.
+ */
+void
+ipset_envopt_set(struct ipset_session *session, enum ipset_envopt opt)
+{
+	assert(session);
+	session->envopts |= opt;
+}
+
+/**
+ * ipset_envopt_unset - unset environment option
+ * @session: session structure
+ * @opt: environment option
+ *
+ * Unset an environment option of the session.
+ */
+void
+ipset_envopt_unset(struct ipset_session *session, enum ipset_envopt opt)
+{
+	assert(session);
+	session->envopts &= ~opt;
+}
+
 /**
  * ipset_session_output - set the session output mode
  * @session: session structure
@@ -211,6 +227,10 @@ ipset_session_report(struct ipset_session *session,
 	assert(session);
 	assert(fmt);
 
+	/* Suppress warning/notice when more important message is required */
+	if (session->err_type > IPSET_NO_ERROR && session->err_type < type)
+		session->report[0] = '\0';
+
 	if (session->lineno != 0 && type == IPSET_ERROR) {
 		sprintf(session->report, "Error in line %u: ",
 			session->lineno);
@@ -228,14 +248,24 @@ ipset_session_report(struct ipset_session *session,
 	if (strlen(session->report) < IPSET_ERRORBUFLEN - 1)
 		strcat(session->report, "\n");
 
-	if (type == IPSET_ERROR) {
-		session->errmsg = session->report;
-		session->warnmsg = NULL;
+	session->err_type = type;
+	if (type == IPSET_ERROR)
 		ipset_data_reset(ipset_session_data(session));
-	} else {
-		session->errmsg = NULL;
-		session->warnmsg = session->report;
-	}
+
+	return -1;
+}
+
+/**
+ * ipset_session_warning_as_error - set warning as error
+ * @session: session structrure
+ *
+ * Returns -1.
+ */
+int
+ipset_session_warning_as_error(struct ipset_session *session)
+{
+	session->err_type = IPSET_ERROR;
+	ipset_data_reset(ipset_session_data(session));
 	return -1;
 }
 
@@ -250,37 +280,36 @@ ipset_session_report_reset(struct ipset_session *session)
 {
 	assert(session);
 	session->report[0] = '\0';
-	session->errmsg = session->warnmsg = NULL;
+	session->err_type = IPSET_NO_ERROR;
 }
 
 /**
- * ipset_session_error - return the report buffer as error
+ * ipset_session_report_msg - return the report buffer
  * @session: session structure
  *
- * Return the pointer to the report buffer as an error report.
- * If there is no error message in the buffer, NULL returned.
+ * Return the pointer to the report buffer.
+ * If there is no error message, the buffer is empty.
  */
 const char *
-ipset_session_error(const struct ipset_session *session)
+ipset_session_report_msg(const struct ipset_session *session)
 {
 	assert(session);
 
-	return session->errmsg;
+	return session->report;
 }
 
 /**
- * ipset_session_warning - return the report buffer as warning
+ * ipset_session_report_type - return the type of the report
  * @session: session structure
  *
- * Return the pointer to the report buffer as a warning report.
- * If there is no warning message in the buffer, NULL returned.
+ * Return the type of the message in the report buffer.
  */
-const char *
-ipset_session_warning(const struct ipset_session *session)
+enum ipset_err_type
+ipset_session_report_type(const struct ipset_session *session)
 {
 	assert(session);
 
-	return session->warnmsg;
+	return session->err_type;
 }
 
 /*
@@ -336,6 +365,10 @@ static const struct ipset_attr_policy cmd_attrs[] = {
 		.type = MNL_TYPE_U32,
 		.opt = IPSET_OPT_LINENO,
 	},
+	[IPSET_ATTR_INDEX] = {
+		.type = MNL_TYPE_U16,
+		.opt = IPSET_OPT_INDEX,
+	},
 };
 
 static const struct ipset_attr_policy create_attrs[] = {
@@ -663,6 +696,10 @@ attr2data(struct ipset_session *session, struct nlattr *nla[],
 		default:
 			break;
 		}
+	} else if (attr->type == MNL_TYPE_NUL_STRING) {
+		if (!d || strlen(d) >= attr->len)
+			FAILURE("Broken kernel message: "
+				"string type attribute missing or too long!");
 	}
 #ifdef IPSET_DEBUG
 	 else
@@ -703,7 +740,8 @@ static const char cmd2name[][9] = {
 static inline int
 call_outfn(struct ipset_session *session)
 {
-	int ret = session->outfn("%s", session->outbuf);
+	int ret = session->print_outfn(session, session->p,
+				      "%s", session->outbuf);
 
 	session->outbuf[0] = '\0';
 
@@ -817,8 +855,9 @@ list_adt(struct ipset_session *session, struct nlattr *nla[])
 	if (session->mode == IPSET_LIST_XML)
 		safe_snprintf(session, "</elem>");
 
-	for (arg = type->args[IPSET_ADD]; arg != NULL && arg->opt; arg++) {
-		D("print arg opt %u %s", arg->opt,
+	for (i = 0; type->cmd[IPSET_ADD].args[i] != IPSET_ARG_NONE; i++) {
+		arg = ipset_keyword(type->cmd[IPSET_ADD].args[i]);
+		D("print arg opt %u (%s) %s", arg->opt, arg->name[0],
 		   ipset_data_test(data, arg->opt) ? "(yes)" : "(missing)");
 		if (!(arg->print && ipset_data_test(data, arg->opt)))
 			continue;
@@ -905,7 +944,12 @@ list_create(struct ipset_session *session, struct nlattr *nla[])
 		break;
 	}
 
-	for (arg = type->args[IPSET_CREATE]; arg != NULL && arg->opt; arg++) {
+	D("type %s, rev %u", type->name, type->revision);
+	for (i = 0; type->cmd[IPSET_CREATE].args[i] != IPSET_ARG_NONE; i++) {
+		arg = ipset_keyword(type->cmd[IPSET_CREATE].args[i]);
+		D("create print arg opt %u (%s) %s", arg->opt,
+		   arg->name[0] ? arg->name[0] : "",
+		   ipset_data_test(data, arg->opt) ? "(yes)" : "(missing)");
 		if (!arg->print ||
 		    !ipset_data_test(data, arg->opt) ||
 		    (arg->opt == IPSET_OPT_FAMILY &&
@@ -1131,6 +1175,7 @@ callback_version(struct ipset_session *session, struct nlattr *nla[])
 			   "while userspace supports protocol versions %u-%u",
 			   min, max, IPSET_PROTOCOL_MIN, IPSET_PROTOCOL_MAX);
 
+	session->protocol = MIN(max, IPSET_PROTOCOL_MAX);
 	session->version_checked = true;
 
 	return MNL_CB_STOP;
@@ -1258,10 +1303,10 @@ callback_data(const struct nlmsghdr *nlh, void *data)
 	proto = mnl_attr_get_u8(nla[IPSET_ATTR_PROTOCOL]);
 
 	/* Check protocol */
-	if (cmd != IPSET_CMD_PROTOCOL && proto != IPSET_PROTOCOL)
+	if (cmd != IPSET_CMD_PROTOCOL && proto != session->protocol)
 		FAILURE("Giving up: kernel protocol version %u "
 			"does not match our protocol version %u",
-			proto, IPSET_PROTOCOL);
+			proto, session->protocol);
 
 	D("Message: %s", cmd2name[cmd]);
 	switch (cmd) {
@@ -1420,8 +1465,8 @@ callback_error(const struct nlmsghdr *nlh, void *cbdata)
 		if (!(session->envopts & IPSET_ENV_QUIET)) {
 			ipset_print_elem(session->report, IPSET_ERRORBUFLEN,
 					 session->data, IPSET_OPT_NONE, 0);
-			ipset_warn(session, " is NOT in set %s.",
-				   ipset_data_setname(data));
+			ipset_notice(session, " is NOT in set %s.",
+				     ipset_data_setname(data));
 		}
 		return ret;
 	}
@@ -1563,8 +1608,8 @@ data2attr(struct ipset_session *session, struct nlmsghdr *nlh,
 			    type, family, attrs);
 }
 
-#define ADDATTR_PROTOCOL(nlh)						\
-	mnl_attr_put_u8(nlh, IPSET_ATTR_PROTOCOL, IPSET_PROTOCOL)
+#define ADDATTR_PROTOCOL(nlh, protocol)					\
+	mnl_attr_put_u8(nlh, IPSET_ATTR_PROTOCOL, protocol)
 
 #define ADDATTR(session, nlh, data, type, family, attrs)		\
 	data2attr(session, nlh, data, type, family, attrs)
@@ -1616,7 +1661,8 @@ build_send_private_msg(struct ipset_session *session, enum ipset_cmd cmd)
 	/* Initialize header */
 	session->transport->fill_hdr(session->handle, cmd, buffer, len, 0);
 
-	ADDATTR_PROTOCOL(nlh);
+	ADDATTR_PROTOCOL(nlh,
+		cmd == IPSET_CMD_PROTOCOL ? IPSET_PROTOCOL : session->protocol);
 
 	switch (cmd) {
 	case IPSET_CMD_PROTOCOL:
@@ -1680,7 +1726,7 @@ build_msg(struct ipset_session *session, bool aggregate)
 					     session->buffer,
 					     session->bufsize,
 					     session->envopts);
-		ADDATTR_PROTOCOL(nlh);
+		ADDATTR_PROTOCOL(nlh, session->protocol);
 	}
 	D("Protocol added, aggregate %s", aggregate ? "yes" : "no");
 	switch (session->cmd) {
@@ -1926,7 +1972,7 @@ ipset_cmd(struct ipset_session *session, enum ipset_cmd cmd, uint32_t lineno)
 
 	assert(session);
 
-	if (cmd <= IPSET_CMD_NONE || cmd >= IPSET_MSG_MAX)
+	if (cmd < IPSET_CMD_NONE || cmd >= IPSET_MSG_MAX)
 		return 0;
 
 	/* Initialize transport method if not done yet */
@@ -1940,7 +1986,14 @@ ipset_cmd(struct ipset_session *session, enum ipset_cmd cmd, uint32_t lineno)
 	if (!session->version_checked) {
 		if (build_send_private_msg(session, IPSET_CMD_PROTOCOL) < 0)
 			return -1;
+		if (ipset_session_report_type(session) == IPSET_WARNING &&
+		    cmd != IPSET_CMD_NONE)
+			/* Suppress protocol warning */
+			ipset_session_report_reset(session);
 	}
+	/* IPSET_CMD_NONE: check protocol version only */
+	if (cmd == IPSET_CMD_NONE)
+		return 0;
 
 	/* Private commands */
 	if (cmd == IPSET_CMD_TYPE || cmd == IPSET_CMD_HEADER)
@@ -2009,29 +2062,57 @@ cleanup:
 	return ret;
 }
 
+static
+int __attribute__ ((format (printf, 3, 4)))
+default_print_outfn(struct ipset_session *session, void *p UNUSED,
+		    const char *fmt, ...)
+{
+	int len;
+	va_list args;
+
+	va_start(args, fmt);
+	len = vfprintf(session->ostream, fmt, args);
+	va_end(args);
+
+	return len;
+}
+
 /**
- * ipset_session_outfn - set session output printing function
+ * ipset_session_print_outfn - set session output printing function
+ * @session: session structure
+ * @outfn: output printing function
+ * @p: pointer to private area
  *
- * Set the session printing function.
+ * Set the session output printing function. If the @outfn is NULL,
+ * then the default output function is configured. You can set
+ * the @p pointer to a private area: the output printing function
+ * is called with @p in one of its arguments.
  *
+ * Returns 0 on success or a negative error code.
  */
 int
-ipset_session_outfn(struct ipset_session *session, ipset_outfn outfn)
+ipset_session_print_outfn(struct ipset_session *session,
+			  ipset_print_outfn outfn,
+			  void *p)
 {
-	session->outfn = outfn ? outfn : printf;
+	session->print_outfn = outfn ? outfn : default_print_outfn;
+	session->p = p;
 	return 0;
 }
 
 /**
  * ipset_session_init - initialize an ipset session
+ * @outfn: output printing function
+ * @p: pointer to private area
  *
  * Initialize an ipset session by allocating a session structure
- * and filling out with the initialization data.
+ * and filling out with the initialization data. The function
+ * calls ipset_session_print_outfn() to set @print_outfn, @p.
  *
  * Returns the created session sctructure on success or NULL.
  */
 struct ipset_session *
-ipset_session_init(ipset_outfn outfn)
+ipset_session_init(ipset_print_outfn print_outfn, void *p)
 {
 	struct ipset_session *session;
 	size_t bufsize = getpagesize();
@@ -2042,12 +2123,15 @@ ipset_session_init(ipset_outfn outfn)
 		return NULL;
 	session->bufsize = bufsize;
 	session->buffer = session + 1;
+	session->istream = stdin;
+	session->ostream = stdout;
+	session->protocol = IPSET_PROTOCOL;
 
 	/* The single transport method yet */
 	session->transport = &ipset_mnl_transport;
 
 	/* Output function */
-	session->outfn = outfn;
+	ipset_session_print_outfn(session, print_outfn, p);
 
 	/* Initialize data structures */
 	session->data = ipset_data_init();
@@ -2062,6 +2146,197 @@ free_session:
 	return NULL;
 }
 
+/**
+ * ipset_session_io_full - set full IO for the session
+ * @session: session structure
+ * @filename: filename
+ * @what: operate on input/output
+ *
+ * The normal "-file" CLI interface does not provide an interface
+ * to set both the input (restore) and output (list/save) for
+ * a session. This function makes it possible to configure those.
+ *
+ * When a filename for input is passed, then the file will be opened
+ * for reading.
+ * When a filename for output is passed, then the file will be opened
+ * for writing.
+ * Previously opened files are closed.
+ * If NULL is passed as filename, stdin/stdout is set.
+ * Input/output files can be set separatedly.
+ * The function returns error if the file cannot be opened or
+ * normal IO mode is already set.
+ *
+ * Returns 0 on success or a negative error code.
+ */
+int
+ipset_session_io_full(struct ipset_session *session, const char *filename,
+		      enum ipset_io_type what)
+{
+	FILE *f;
+
+	assert(session);
+
+	if (session->normal_io)
+		return ipset_err(session,
+			"Normal IO is in use, full IO cannot be selected");
+
+	switch (what) {
+	case IPSET_IO_INPUT:
+		if (session->istream != stdin)
+			fclose(session->istream);
+		if (!filename) {
+			session->istream = stdin;
+		} else {
+			f = fopen(filename, "r");
+			if (!f)
+				return ipset_err(session,
+					"Cannot open %s for reading: %s",
+					filename, strerror(errno));
+			session->istream = f;
+		}
+		break;
+	case IPSET_IO_OUTPUT:
+		if (session->ostream != stdout)
+			fclose(session->ostream);
+		if (!filename) {
+			session->ostream = stdout;
+		} else {
+			f = fopen(filename, "w");
+			if (!f)
+				return ipset_err(session,
+					"Cannot open %s for writing: %s",
+					filename, strerror(errno));
+			session->ostream = f;
+		}
+		break;
+	default:
+		return ipset_err(session,
+				"Library error, invalid ipset_io_type");
+	}
+	session->full_io = !(session->istream == stdin &&
+			     session->ostream == stdout);
+	return 0;
+}
+
+/**
+ * ipset_session_io_normal - set normal IO for the session
+ * @session: session structure
+ * @filename: filename
+ * @what: operate on input/output
+ *
+ * The normal "-file" CLI interface to set either the input (restore)
+ * or output (list/save) for a session. This function does not make
+ * possible to set both independently.
+ *
+ * When a filename for input is passed, then the file will be opened
+ * for reading.
+ * When a filename for output is passed, then the file will be opened
+ * for writing.
+ * Previously opened files are closed.
+ * If NULL is passed as filename, stdin/stdout is set.
+ * Input/output files cannot be set separatedly.
+ * The function returns error if the file cannot be opened or
+ * full IO mode is already set.
+ *
+ * Returns 0 on success or a negative error code.
+ */
+int
+ipset_session_io_normal(struct ipset_session *session, const char *filename,
+			enum ipset_io_type what)
+{
+	FILE *f;
+
+	assert(session);
+	assert(filename);
+
+	if (session->full_io)
+		return ipset_err(session,
+			"Full IO is in use, normal IO cannot be selected");
+	if (session->istream != stdin) {
+		fclose(session->istream);
+		session->istream = stdin;
+	}
+	if (session->ostream != stdout) {
+		fclose(session->ostream);
+		session->ostream = stdout;
+	}
+	switch (what) {
+	case IPSET_IO_INPUT:
+		f = fopen(filename, "r");
+		if (!f)
+			return ipset_err(session,
+				"Cannot open %s for reading: %s",
+				filename, strerror(errno));
+		session->istream = f;
+		break;
+	case IPSET_IO_OUTPUT:
+		f = fopen(filename, "w");
+		if (!f)
+			return ipset_err(session,
+				"Cannot open %s for writing: %s",
+				filename, strerror(errno));
+		session->ostream = f;
+		break;
+	default:
+		return ipset_err(session,
+				"Library error, invalid ipset_io_type");
+	}
+	session->normal_io = !(session->istream == stdin &&
+			       session->ostream == stdout);
+	return 0;
+}
+
+/**
+ * ipset_session_io_stream - returns the input or output stream
+ * @what: operate on input/output
+ *
+ * Returns the input or output stream of the session.
+ */
+FILE *
+ipset_session_io_stream(struct ipset_session *session,
+			enum ipset_io_type what)
+{
+	switch (what) {
+	case IPSET_IO_INPUT:
+		return session->istream;
+	case IPSET_IO_OUTPUT:
+		return session->ostream;
+	default:
+		return NULL;
+	}
+}
+
+/**
+ * ipset_session_io_close - closes the input or output stream
+ * @what: operate on input/output
+ *
+ * Closes the input or output stream of the session.
+ *
+ * Returns 0 on success or a negative error code.
+ */
+int
+ipset_session_io_close(struct ipset_session *session,
+		       enum ipset_io_type what)
+{
+	switch (what) {
+	case IPSET_IO_INPUT:
+		if (session->istream != stdin) {
+			fclose(session->istream);
+			session->istream = stdin;
+		}
+		break;
+	case IPSET_IO_OUTPUT:
+		if (session->ostream != stdout) {
+			fclose(session->ostream);
+			session->ostream = stdout;
+		}
+		break;
+	default:
+		break;
+	}
+	return 0;
+}
+
 /**
  * ipset_session_fini - destroy an ipset session
  * @session: session structure
@@ -2079,6 +2354,10 @@ ipset_session_fini(struct ipset_session *session)
 		session->transport->fini(session->handle);
 	if (session->data)
 		ipset_data_fini(session->data);
+	if (session->istream != stdin)
+		fclose(session->istream);
+	if (session->ostream != stdout)
+		fclose(session->ostream);
 
 	ipset_cache_fini();
 	free(session);
